/***************************************************************************

    M.A.M.E.CE  -  Multiple Arcade Machine Emulator for WinCE
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse
	WinCE Portions Copyright (C) 1999 Benjamin Cooley
    
    This file is part of MAMECE, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************

  CESound.c

 ***************************************************************************/

#include "mame.h"
#include "osdepend.h"
#include "mamece.h"
#include "CESound.h"

/***************************************************************************
    function prototypes
 ***************************************************************************/

static int      CESound_init(options_type *options);
static void     CESound_exit(void);
static void     CESound_update_audio(void);
static void     CESound_play_sample(int channel, signed char* data, int len, int freq, int volume, int loop);
static void     CESound_play_sample_16(int channel, signed short* data, int len, int freq, int volume, int loop);
static void     CESound_play_streamed_sample(int channel, signed char* data, int len, int freq, int volume, int pan);
static void     CESound_play_streamed_sample_16(int channel, signed short* data, int len, int freq, int volume, int pan);
static void     CESound_set_sample_freq(int channel,int freq);
static void     CESound_set_sample_volume(int channel,int volume);
static void     CESound_stop_sample(int channel);
static void     CESound_restart_sample(int channel);
static int      CESound_get_sample_status(int channel);
static void     CESound_set_mastervolume(int volume);
static int      CESound_get_mastervolume();
static void     CESound_sound_enable(int enable);

/***************************************************************************
    External variables
 ***************************************************************************/

struct OSDSound CESound = 
{
    { CESound_init },                     /* init                    */
    { CESound_exit },                     /* exit                    */
    { CESound_update_audio },             /* update_audio            */
    { CESound_play_sample },              /* play_sample             */
    { CESound_play_sample_16 },           /* play_sample_16          */
    { CESound_play_streamed_sample },     /* play_streamed_sample    */
    { CESound_play_streamed_sample_16 },  /* play_streamed_sample_16 */
    { CESound_set_sample_freq },          /* adjust_sample_freq      */
    { CESound_set_sample_volume },        /* adjust_sample_volume    */
    { CESound_stop_sample },              /* stop_sample             */
    { CESound_restart_sample },           /* restart_sample          */
    { CESound_get_sample_status },        /* get_sample_status       */
    { CESound_set_mastervolume },         /* set_mastervolume        */
    { CESound_get_mastervolume },         /* get_mastervolume        */
    { CESound_sound_enable }              /* sound_enable            */
};

/***************************************************************************
    Internal structures
 ***************************************************************************/

#define NUM_WAVEHDRS  4

struct tSound_private
{
	WAVEOUTCAPS		m_Caps;
	HWAVEOUT		m_hWaveOut;
	WAVEHDR			m_WaveHdrs[NUM_WAVEHDRS];

	int				m_nSampleRate;
	int				m_nSampleBits;
	int				m_nChannels;
	int				m_nVolume;		// -32 to 0 attenuation value
};

/***************************************************************************
    Internal variables
 ***************************************************************************/

static struct tSound_private      This;

/***************************************************************************
    External OSD functions  
 ***************************************************************************/

static int CESound_init(options_type *options)
{
	MMRESULT res;
	WAVEFORMATEX wf;

	This.m_hWaveOut = NULL;

	if (Machine->sample_rate == 0)
	{
		Machine->sample_rate = options->sample_rate;
		Machine->sample_bits = options->sample_bits;
	}

	This.m_nSampleRate = Machine->sample_rate;
	This.m_nSampleBits = Machine->sample_bits;
	This.m_nChannels = 1;

	wf.wFormatTag = WAVE_FORMAT_PCM;
	wf.nChannels = This.m_nChannels; 
	wf.nSamplesPerSec = This.m_nSampleRate; 
	wf.nAvgBytesPerSec = This.m_nSampleRate * This.m_nSampleBits / 8; 
	wf.nBlockAlign = This.m_nSampleBits * This.m_nChannels / 8;
	wf.wBitsPerSample = This.m_nSampleBits; 
	wf.cbSize = 0;

	res = waveOutOpen(
		&This.m_hWaveOut,	// Handle
		0,					// ID (0 for wave mapper)
		&wf,				// Wave format
		0,					// Callback
		0,					// Instance data
		CALLBACK_NULL);

	if (res != MMSYSERR_NOERROR)
		return 1;

	memset(&This.m_WaveHdrs, 0, sizeof(WAVEHDR) * NUM_WAVEHDRS);

	This.m_nVolume = 0;

	return 0;
}

static void CESound_exit(void)
{
	BOOL done;
	int i, ticks;

	ticks = GetTickCount();
	done = FALSE;
	while (!done && GetTickCount() - ticks < 500)
	{
		done = TRUE;
		for (i = 0; i < NUM_WAVEHDRS; i++)
		{
			if (This.m_WaveHdrs[i].dwFlags & WHDR_DONE)
			{
				waveOutUnprepareHeader(This.m_hWaveOut, &This.m_WaveHdrs[i], sizeof(WAVEHDR));
				This.m_WaveHdrs[i].lpData = NULL;
				This.m_WaveHdrs[i].dwBufferLength = 0;
				This.m_WaveHdrs[i].dwFlags = 0;
			}
			else if (This.m_WaveHdrs[i].dwFlags)
				done = FALSE;
		}
	}

	waveOutClose(This.m_hWaveOut);
}

static void CESound_update_audio(void)
{
}

static void CESound_play_sample(int channel, signed char *data, int len, int freq, int volume, int loop)
{
}

static void CESound_play_sample_16(int channel, signed short *data, int len, int freq, int volume, int loop)
{
	int i;
	signed short *s;
	signed short *d;
	DWORD pos, add, end;

	for (i = 0; i < NUM_WAVEHDRS; i++)
	{
		if (This.m_WaveHdrs[i].dwFlags == 0 || This.m_WaveHdrs[i].dwFlags & WHDR_DONE)
		{
			int newlen = len * This.m_nSampleRate / freq;
			if (This.m_WaveHdrs[i].lpData && This.m_WaveHdrs[i].dwBufferLength != newlen)
			{
				waveOutUnprepareHeader(This.m_hWaveOut, &This.m_WaveHdrs[i], sizeof(WAVEHDR));
				free(This.m_WaveHdrs[i].lpData);
				This.m_WaveHdrs[i].lpData = NULL;
			}
			if (!This.m_WaveHdrs[i].lpData)
			{
				This.m_WaveHdrs[i].lpData = (LPSTR)malloc(newlen << 1);
				This.m_WaveHdrs[i].dwBufferLength = newlen;
				This.m_WaveHdrs[i].dwFlags = 0;
				waveOutPrepareHeader(This.m_hWaveOut, &This.m_WaveHdrs[i], sizeof(WAVEHDR));
			}
			s = data;
			d = (signed short *)This.m_WaveHdrs[i].lpData;
			add = (((DWORD)(freq << 15) / (DWORD)This.m_nSampleRate) << 1) + 3;
			end = len << 16;
			for (pos = 0; pos < end; pos += add)
				*d++ = (signed char)s[(pos >> 16)];
			waveOutWrite(This.m_hWaveOut, &This.m_WaveHdrs[i], sizeof(WAVEHDR));
			return;	
		}
	}
}

static void CESound_play_streamed_sample(int channel, signed char *data, int len, int freq, int volume, int pan)
{
	int i;
	signed char *s;
	unsigned char *d;
	DWORD pos, add, end;
	int rate;

	rate = This.m_nSampleRate * 100 / 75;

	for (i = 0; i < NUM_WAVEHDRS; i++)
	{
		if (This.m_WaveHdrs[i].dwFlags == 0 || This.m_WaveHdrs[i].dwFlags & WHDR_DONE)
		{
			int newlen = len * rate / freq;
			if (This.m_WaveHdrs[i].lpData && This.m_WaveHdrs[i].dwBufferLength != newlen)
			{
				waveOutUnprepareHeader(This.m_hWaveOut, &This.m_WaveHdrs[i], sizeof(WAVEHDR));
				free(This.m_WaveHdrs[i].lpData);
				This.m_WaveHdrs[i].lpData = NULL;
			}
			if (!This.m_WaveHdrs[i].lpData)
			{
				This.m_WaveHdrs[i].lpData = (LPSTR)malloc(newlen);
				This.m_WaveHdrs[i].dwBufferLength = newlen;
				This.m_WaveHdrs[i].dwFlags = 0;
				waveOutPrepareHeader(This.m_hWaveOut, &This.m_WaveHdrs[i], sizeof(WAVEHDR));
			}
			s = data;
			d = (unsigned char *)This.m_WaveHdrs[i].lpData;
			add = (((DWORD)(freq << 15) / (DWORD)rate) << 1) + 3;
			end = len << 16;
			for (pos = 0; pos < end; pos += add)
				*d++ = (unsigned char)((int)s[(pos >> 16)] + 128);
			waveOutWrite(This.m_hWaveOut, &This.m_WaveHdrs[i], sizeof(WAVEHDR));
			return;	
		}
	}
}

static void CESound_play_streamed_sample_16(int channel, signed short *data, int len, int freq, int volume, int pan)
{
}

static void CESound_set_sample_freq(int channel,int freq)
{
}

static void CESound_set_sample_volume(int channel,int volume)
{
}

static void CESound_stop_sample(int channel)
{
}

static void CESound_restart_sample(int channel)
{
}

static int CESound_get_sample_status(int channel)
{
    return 0;
}

static void CESound_set_mastervolume(int volume)
{
	This.m_nVolume = volume;
}

static int CESound_get_mastervolume()
{
    return This.m_nVolume;
}

static void CESound_sound_enable(int enable)
{

}

/***************************************************************************
    Internal functions
 ***************************************************************************/

