/***************************************************************************

    M.A.M.E.CE  -  Multiple Arcade Machine Emulator for WinCE
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse
	WinCE Portions Copyright (C) 1999 Benjamin Cooley
    
    This file is part of MAMECE, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************

  Cgamex.c

 ***************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <gamex.h>

#ifndef VK_0
#define VK_0 0x30
#define VK_1 0x31
#define VK_2 0x32
#define VK_3 0x33
#define VK_4 0x34
#define VK_5 0x35
#define VK_6 0x36
#define VK_7 0x37
#define VK_8 0x38
#define VK_9 0x39
#endif

#define VK_CASSIOPEIA_SLIDER	0x86
#define VK_CASSIOPEIA_ESC		0x87
#define VK_CASSIOPEIA_REC		0xC1
#define VK_CASSIOPEIA_B1		0xC2
#define VK_CASSIOPEIA_B2		0xC3
#define VK_CASSIOPEIA_B3		0xC4

/***************************************************************************
    Function prototypes
 ***************************************************************************/

extern "C" {

BOOL		    GAMEX_Init();
void			GAMEX_Close();
void*			GAMEX_GetFBAddress();
int				GAMEX_GetFBModulo();
int				GAMEX_GetFBBpp();
BOOL			GAMEX_BeginDraw();
BOOL			GAMEX_EndDraw();
BOOL			GAMEX_GetScreenRect(RECT *prc);
SHORT			GAMEX_GetAsyncKeyState(int key);
}

/***************************************************************************
    Internal structures
 ***************************************************************************/

/***************************************************************************
    Internal variables
 ***************************************************************************/

/* our dialog/configured options */
static GameX *gpgmx;

/***************************************************************************
    Internal functions
 ***************************************************************************/

/***************************************************************************
    External functions  
 ***************************************************************************/

BOOL GAMEX_Init()
{
	gpgmx = new GameX();

	if (!gpgmx->OpenGraphics() || !gpgmx->OpenButtons(NULL))
	{
		delete gpgmx;
		return FALSE;
	}

	return TRUE;
}

void GAMEX_Close()
{
	delete gpgmx;
}

void* GAMEX_GetFBAddress()
{
	return gpgmx->GetFBAddress();
}

int	GAMEX_GetFBModulo()
{
	return gpgmx->GetFBModulo();
}

int	GAMEX_GetFBBpp()
{
	return gpgmx->GetFBBpp();
}

BOOL GAMEX_BeginDraw()
{
	return gpgmx->BeginDraw();
}

BOOL GAMEX_EndDraw()
{
	return gpgmx->EndDraw();
}

BOOL GAMEX_GetScreenRect(RECT *prc)
{
	return gpgmx->GetScreenRect(prc);
}

SHORT GAMEX_GetAsyncKeyState(int key)
{
	int newkey = key;

#ifdef CASSIOPEIA_E100
	switch (key)
	{
	  case VK_1:			newkey = VK_CASSIOPEIA_B1; break;
	  case VK_2:			newkey = VK_CASSIOPEIA_B2; break;
	  case VK_3:			newkey = VK_CASSIOPEIA_SLIDER; break;
	  case VK_ESCAPE:		newkey = VK_CASSIOPEIA_ESC; break;
	  case VK_TAB:			newkey = VK_CASSIOPEIA_REC; break;
	  case VK_LCONTROL:		newkey = VK_CASSIOPEIA_B1; break;
	  case VK_RETURN:		newkey = VK_CASSIOPEIA_B1; break;
	  case VK_MENU:			newkey = VK_CASSIOPEIA_B2; break;
	  case VK_SPACE:		newkey = VK_CASSIOPEIA_B3; break;
	}
#endif

	return GetAsyncKeyState(newkey);
}
