/***************************************************************************

    M.A.M.E.CE  -  Multiple Arcade Machine Emulator for WinCE
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse
	WinCE Portions Copyright (C) 1999 Benjamin Cooley
    
    This file is part of MAMECE, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************

  GDIDisplay.c

 ***************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <assert.h>
#include <math.h>
#include "mamece.h"
#include "resource.h"
#include "driver.h"
#include "GDIDisplay.h"
//#include "M32Util.h"
//#include "status.h"
#include "dirty.h"

#define MAKECOL(r, g, b) (((r) & 0x001F) << 10 | ((g) & 0x001F) << 5 | ((b) & 0x001F))

struct GDIBitmap
{
	int linesize;
	int bufsize;
	HBITMAP bitmap;
	BITMAPINFO bi;
	// Don't add anything after BITMAPINFO.. as the RGBQUAD colors must follow directly after it!!
};

/***************************************************************************
    function prototypes
 ***************************************************************************/

static void             OnCommand(HWND hWnd, int CmdID, HWND hWndCtl, UINT codeNotify);
static void             OnPaint(HWND hWnd);
static void             OnPaletteChanged(HWND hWnd, HWND hWndPaletteChange);
static BOOL             OnQueryNewPalette(HWND hWnd);
//static void             OnGetMinMaxInfo(HWND hWnd, MINMAXINFO* pMinMaxInfo);
static BOOL             OnEraseBkgnd(HWND hWnd, HDC hDC);
static void             OnSize(HWND hwnd,UINT state,int cx,int cy);
static void             OnDrawItem(HWND hwnd, const DRAWITEMSTRUCT *lpDrawItem);
static struct			osd_bitmap* AllocBitmap(struct osd_bitmap *bm, 
							int width, int height, int depth, LPPALETTEENTRY lpPal);
static void             SetPen(int pen, unsigned char red, unsigned char green, unsigned char blue);
static void             SetPaletteColors(void);
static LRESULT CALLBACK AboutDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static void             AdjustVisibleRect(int xmin, int ymin, int xmax, int ymax);
static void             AdjustPalette(void);

static int                GDI_init(options_type *options);
static void               GDI_exit(void);
static struct osd_bitmap* GDI_new_bitmap(int width, int height,int depth);
static void               GDI_clearbitmap(struct osd_bitmap* bitmap);
static void               GDI_free_bitmap(struct osd_bitmap* bitmap);
static struct osd_bitmap* GDI_create_display(int width, int height, int attributes);
static int                GDI_set_display(int width, int height, int attributes);
static void               GDI_close_display(void);
static void               GDI_allocate_colors(unsigned int totalcolors, const unsigned char *palette, unsigned short *pens);
static void               GDI_modify_pen(int pen, unsigned char red, unsigned char green, unsigned char blue);
static void               GDI_get_pen(int pen, unsigned char* pRed, unsigned char* pGreen, unsigned char* pBlue);
static void               GDI_mark_dirty(int x1, int y1, int x2, int y2, int ui);
static void               GDI_update_display(void);
static void               GDI_led_w(int led, int on);
static void               GDI_set_gamma(float gamma);
static void               GDI_set_brightness(int brightness);
static void               GDI_save_snapshot(void);
static BOOL               GDI_OnMessage(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam, LRESULT* pResult);
static void               GDI_Refresh(void);
static int                GDI_GetBlackPen(void);
static void               GDI_UpdateFPS(BOOL bShow, int nSpeed, int nFPS, int nMachineFPS, int nFrameskip, int nVecUPS);

/***************************************************************************
    External variables
 ***************************************************************************/

struct OSDDisplay   GDIDisplay = 
{
    { GDI_init },               /* init              */
    { GDI_exit },               /* exit              */
    { GDI_new_bitmap },         /* new_bitmap        */
    { GDI_clearbitmap },        /* clearbitmap       */
    { GDI_free_bitmap },        /* free_bitmap       */
    { GDI_create_display },     /* create_display    */
    { GDI_set_display },        /* set_display       */
    { GDI_close_display },      /* close_display     */
    { GDI_allocate_colors },    /* allocate_colors   */
    { GDI_modify_pen },         /* modify_pen        */
    { GDI_get_pen },            /* get_pen           */
    { GDI_mark_dirty },         /* mark_dirty        */
    { 0 },                      /* skip_this_frame   */
    { GDI_update_display },     /* update_display    */
    { GDI_led_w },              /* led_w             */
    { GDI_set_gamma },          /* set_gamma         */
    { Display_get_gamma },      /* get_gamma         */
    { GDI_set_brightness },     /* set_brightness    */
    { Display_get_brightness }, /* get_brightness    */
    { GDI_save_snapshot },      /* save_snapshot     */

    { GDI_OnMessage },          /* OnMessage         */
    { GDI_Refresh },            /* Refresh           */
    { GDI_GetBlackPen },        /* GetBlackPen       */
    { GDI_UpdateFPS },          /* UpdateFPS         */
};

/***************************************************************************
    Internal structures
 ***************************************************************************/

struct tDisplay_private
{
    struct osd_bitmap*  m_pMAMEBitmap;
    struct osd_bitmap*  m_pUpdateBitmap;

	BYTE**				m_pScreenLines;

    RECT                m_ClientRect;
    RECT                m_DrawRect;

//    int                 m_nClientWidth;
//    int                 m_nClientHeight;
//    int                 m_nWindowWidth;
//    int                 m_nWindowHeight;

	BOOL				m_bStretch;
    BOOL                m_bScanlines;
    BOOL                m_bDouble;
    BOOL                m_bVectorDouble;
    BOOL                m_bUseDirty;
	int					m_nScreenWidth;
	int					m_nScreenHeight;
	int					m_nScreenDepth;
	int                 m_nDepth;

    /* AVI Capture params */
    BOOL                m_bAviCapture;      /* Capture AVI mode */
    BOOL                m_bAviRun;          /* Capturing frames */
    int                 m_nAviShowMessage;  /* Show status if !0 */

    BOOL                m_bUpdatePalette;
    HPALETTE            m_hPalette;
    PALETTEENTRY        m_PalEntries[OSD_NUMPENS];
    PALETTEENTRY        m_AdjustedPalette[OSD_NUMPENS];
	WORD				m_Pal16[OSD_NUMPENS];

    UINT                m_nUIPen;

    HWND                m_hWndAbout;
};

/***************************************************************************
    Internal variables
 ***************************************************************************/

static struct tDisplay_private      This;

/***************************************************************************
    External OSD functions  
 ***************************************************************************/

/*
    put here anything you need to do when the program is started. Return 0 if 
    initialization was successful, nonzero otherwise.
*/
static int GDI_init(options_type *options)
{
    OSDDisplay.init(options);

    This.m_pMAMEBitmap      = NULL;
	This.m_pUpdateBitmap    = NULL;

//    memset(&This.m_VisibleRect, 0, sizeof(tRect));
//    This.m_nClientWidth     = 0;
//    This.m_nClientHeight    = 0;
//    This.m_nWindowWidth     = 0;
//    This.m_nWindowHeight    = 0;

	This.m_pScreenLines		= NULL;

	This.m_bStretch         = FALSE; /* options->stretch */	
    This.m_bScanlines       = FALSE; /* options->hscan_lines; */
    This.m_bDouble          = FALSE; /* (options->scale == 2); */
    This.m_bVectorDouble    = FALSE; /* options->double_vector; */
    This.m_bUseDirty        = FALSE; /* options->use_dirty; */
    This.m_nDepth           = 16; /* options->depth; */

    This.m_bUpdatePalette   = FALSE;
    This.m_hPalette         = NULL;
    This.m_hWndAbout        = NULL;
    memset(This.m_PalEntries,      0, OSD_NUMPENS * sizeof(PALETTEENTRY));
    memset(This.m_AdjustedPalette, 0, OSD_NUMPENS * sizeof(PALETTEENTRY));
	memset(This.m_Pal16,		   0, OSD_NUMPENS * sizeof(WORD));
    This.m_nUIPen           = 0;

    /* Scanlines don't work. */
    This.m_bScanlines = FALSE;

    /* Check for inconsistent parameters. */
    if (This.m_bScanlines == TRUE
    &&  This.m_bDouble    == FALSE)
    {
        This.m_bScanlines = FALSE;
    }

    if (This.m_nDepth < 8)
        This.m_nDepth = 8;
    else
    if (8 < This.m_nDepth)
        This.m_nDepth = 16;

    return 0;
}

/*
    put here cleanup routines to be executed when the program is terminated.
*/
static void GDI_exit(void)
{
    OSDDisplay.exit();
}

static struct osd_bitmap* GDI_new_bitmap(int width, int height,int depth)
{
    assert(OSDDisplay.new_bitmap != 0);

    return OSDDisplay.new_bitmap(width, height, depth);
}

static void GDI_clearbitmap(struct osd_bitmap* bitmap)
{
    assert(OSDDisplay.clearbitmap != 0);

    OSDDisplay.clearbitmap(bitmap);
}

static void GDI_free_bitmap(struct osd_bitmap* pBitmap)
{
    assert(OSDDisplay.free_bitmap != 0);

    OSDDisplay.free_bitmap(pBitmap);
}

/*
    Create a display screen, or window, large enough to accomodate a bitmap
    of the given dimensions.
    Return a osd_bitmap pointer or 0 in case of error.
*/
static struct osd_bitmap *GDI_create_display(int width, int height, int attributes)
{
    unsigned int    i;
    RECT            Rect;
    HMENU           hMenu;
    LOGPALETTE      LogPalette;
	DWORD			pal[255];		// Extra space for LOGPALETTE.. must follow LogPalette
    char			TitleBuf[256];
	HDC				hDC;
	HBITMAP			bm;

    /* Find out if display depth is 15 bpp or greater. */
    hDC = GetDC(NULL);
    This.m_nScreenDepth = GetDeviceCaps(hDC, BITSPIXEL) * GetDeviceCaps(hDC, PLANES);
    ReleaseDC(NULL, hDC);

	/* Set up direct access array */
	This.m_nScreenWidth = GetSystemMetrics(SM_CXSCREEN);
	This.m_nScreenHeight = GetSystemMetrics(SM_CYSCREEN);

#ifndef x86
	This.m_pScreenLines = (BYTE**)malloc(This.m_nScreenHeight * sizeof(BYTE*));
	for (i = 0; i < This.m_nScreenHeight; i++)
		This.m_pScreenLines[i] = 0x06140000 + i * 512;
#endif

    if (This.m_nDepth == 16 && 15 <= This.m_nScreenDepth && (attributes & VIDEO_SUPPORTS_16BIT))
        This.m_nDepth = 16;
    else
        This.m_nDepth = 8;

    if (This.m_bUseDirty == TRUE)
    {
        if (Machine->orientation & ORIENTATION_SWAP_XY)
            InitDirty(height, width, NO_DIRTY);
        else
            InitDirty(width, height, NO_DIRTY);
    }

    /* osd_create_bitmap will swap width and height if neccesary. */
    This.m_pMAMEBitmap = osd_new_bitmap(width, height, This.m_nDepth);
    This.m_pUpdateBitmap = osd_new_bitmap(width, height, This.m_nDepth);

    /* Palette */
    LogPalette.palVersion    = 0x0300;
    LogPalette.palNumEntries = OSD_NUMPENS;
    This.m_hPalette = CreatePalette(&LogPalette);

    sprintf(TitleBuf, "%s - %s", Machine->gamedrv->description, WC(MAME32App.m_Name));
    SetWindowText(MAME32App.m_hWnd, WC(TitleBuf));

    GetClientRect(MAME32App.m_hWnd, &This.m_ClientRect);
	
	This.m_DrawRect = This.m_ClientRect;

    /* Fit to screen with correct aspect ratio */
   	if (!This.m_bStretch)
	{
		This.m_DrawRect.right = This.m_pMAMEBitmap->width;
	 	This.m_DrawRect.bottom = This.m_pMAMEBitmap->height;

		if (This.m_pMAMEBitmap->width > This.m_ClientRect.right)
		{
			This.m_DrawRect.right = This.m_ClientRect.right;
			This.m_DrawRect.bottom = This.m_ClientRect.bottom *
				This.m_ClientRect.right / This.m_pMAMEBitmap->width;
		}
		
		if (This.m_pMAMEBitmap->height > This.m_ClientRect.bottom)
		{
			This.m_DrawRect.bottom = This.m_ClientRect.bottom;
			This.m_DrawRect.right = This.m_ClientRect.right *
				This.m_ClientRect.bottom / This.m_pMAMEBitmap->height;
		}

		if (This.m_DrawRect.right < This.m_ClientRect.right)
		{
			This.m_DrawRect.left = (This.m_ClientRect.right - This.m_DrawRect.right) / 2;
			This.m_DrawRect.right += This.m_DrawRect.left;
		}

		if (This.m_DrawRect.bottom < This.m_ClientRect.bottom)
		{
			This.m_DrawRect.top = (This.m_ClientRect.bottom - This.m_DrawRect.bottom) / 2;
			This.m_DrawRect.bottom += This.m_DrawRect.top;
		}
	}

    ShowWindow(MAME32App.m_hWnd, SW_SHOW);
    SetForegroundWindow(MAME32App.m_hWnd);
    InvalidateRect(MAME32App.m_hWnd, NULL, TRUE);
	UpdateWindow(MAME32App.m_hWnd);

    set_ui_visarea(0, 0, This.m_pMAMEBitmap->width - 1, This.m_pMAMEBitmap->height - 1);

    return This.m_pMAMEBitmap;
}

/*
    Set the actual display screen but don't allocate the screen bitmap.
*/
static int GDI_set_display(int width, int height, int attributes)
{
    SetForegroundWindow(MAME32App.m_hWnd);
    return 0; 
}

/*
     Shut down the display
*/
static void GDI_close_display(void)
{
    ExitDirty();
    
    if (This.m_nDepth == 8)
    {
        if (This.m_hPalette != NULL)
        {
            DeletePalette(This.m_hPalette);
            This.m_hPalette = NULL;
        }
    }

    if (IsWindow(This.m_hWndAbout))
    {
        DestroyWindow(This.m_hWndAbout);
        This.m_hWndAbout = NULL;
    }

//    StatusDelete();
   
	SelectObject(This.m_hMemDC, (HGDIOBJ)This.m_hOldBitmap);
    DeleteObject((HGDIOBJ)This.m_hMemDC);

    This.m_pGDIBitmap = NULL;

    osd_free_bitmap(This.m_pMAMEBitmap);
}

/*
    palette is an array of 'totalcolors' R,G,B triplets. The function returns
    in *pens the pen values corresponding to the requested colors.
    If 'totalcolors' is 32768, 'palette' is ignored and the *pens array is filled
    with pen values corresponding to a 5-5-5 15-bit palette
*/
static void GDI_allocate_colors(unsigned int totalcolors, const unsigned char *palette, unsigned short *pens)
{
    unsigned int i;

    if (This.m_nDepth == 8)
    {
        if (totalcolors >= OSD_NUMPENS - 1)
        {
            int bestblack;
            int bestwhite;
            int bestblackscore;
            int bestwhitescore;

            bestblack = bestwhite = 0;
            bestblackscore = 3 * 255 * 255;
            bestwhitescore = 0;
            for (i = 0; i < totalcolors; i++)
            {
                int r, g, b, score;

                r = palette[3 * i];
                g = palette[3 * i + 1];
                b = palette[3 * i + 2];
                score = r * r + g * g + b * b;

                if (score < bestblackscore)
                {
                    bestblack      = i;
                    bestblackscore = score;
                }
                if (score > bestwhitescore)
                {
                    bestwhite      = i;
                    bestwhitescore = score;
                }
            }

            for (i = 0; i < totalcolors; i++)
                pens[i] = i;

            /* map black to pen 0, otherwise the screen border will not be black */
            pens[bestblack] = 0;
            pens[0] = bestblack;

            Machine->uifont->colortable[0] = pens[bestblack];
            Machine->uifont->colortable[1] = pens[bestwhite];
            Machine->uifont->colortable[2] = pens[bestwhite];
            Machine->uifont->colortable[3] = pens[bestblack];

            This.m_nUIPen = pens[bestwhite];
        }
        else
        {
            /*
                If we have free places, fill the palette starting from the end,
                so we don't touch color 0, which is better left black.
            */
            for (i = 0; i < totalcolors; i++)
                pens[i] = OSD_NUMPENS - 1 - i;

            /* As long as there are free pens, set 0 as black for GetBlackPen. */
            SetPen(0, 0, 0, 0);

            /* reserve color 1 for the user interface text */
            This.m_nUIPen = 1;
            SetPen(This.m_nUIPen, 0xFF, 0xFF, 0xFF);

            Machine->uifont->colortable[0] = 0;
            Machine->uifont->colortable[1] = This.m_nUIPen;
            Machine->uifont->colortable[2] = This.m_nUIPen;
            Machine->uifont->colortable[3] = 0;
        }

        for (i = 0; i < totalcolors; i++)
        {
            SetPen(pens[i],
                   palette[3 * i],
                   palette[3 * i + 1],
                   palette[3 * i + 2]);
        }

        SetPaletteColors();
    }
    else
    {
        int r, g, b;

        for (r = 0; r < 32; r++)
            for (g = 0; g < 32; g++)
                for (b = 0; b < 32; b++)
                {
                    int r1, g1, b1;

                    r1 = (int)(31.0 * Display_get_brightness() * pow(r / 31.0, 1.0 / Display_get_gamma()) / 100.0);
                    g1 = (int)(31.0 * Display_get_brightness() * pow(g / 31.0, 1.0 / Display_get_gamma()) / 100.0);
                    b1 = (int)(31.0 * Display_get_brightness() * pow(b / 31.0, 1.0 / Display_get_gamma()) / 100.0);

                    *pens++ = MAKECOL(r1, g1, b1);
                }

        Machine->uifont->colortable[0] = 0;
        Machine->uifont->colortable[1] = MAKECOL(0xFF, 0xFF, 0xFF);
        Machine->uifont->colortable[2] = MAKECOL(0xFF, 0xFF, 0xFF);
        Machine->uifont->colortable[3] = 0;
    }

}

/*
    Change the color of the pen.
*/
static void GDI_modify_pen(int pen, unsigned char red, unsigned char green, unsigned char blue)
{
   assert(This.m_nDepth == 8); /* Shouldn't modify pen in 16 bit color depth game mode! */

   assert(0 <= pen && pen < OSD_NUMPENS);

    if (This.m_PalEntries[pen].peRed    == red
    &&  This.m_PalEntries[pen].peGreen  == green
    &&  This.m_PalEntries[pen].peBlue   == blue)
        return;

    SetPen(pen, red, green, blue);

    This.m_bUpdatePalette = TRUE;
}

/*
    Get the color of a pen.
*/
static void GDI_get_pen(int pen, unsigned char* pRed, unsigned char* pGreen, unsigned char* pBlue)
{
    if (This.m_nDepth == 8)
    {
        assert(0 <= pen && pen < OSD_NUMPENS);

        if (OSD_NUMPENS <= pen)
            pen = 0;

        *pRed   = This.m_PalEntries[pen].peRed;
        *pGreen = This.m_PalEntries[pen].peGreen;
        *pBlue  = This.m_PalEntries[pen].peBlue;
    }
    else
    {
        *pRed   = (pen >> 7) & 0xF8;
        *pGreen = (pen >> 2) & 0xF8;
        *pBlue  = (pen << 3) & 0xF8;
    }
}

static void GDI_mark_dirty(int x1, int y1, int x2, int y2, int ui)
{
}

/*
    Update the display.
*/
static void GDI_update_display(void)
{
    if (This.m_bUpdatePalette == TRUE)
    {
        SetPaletteColors();
        This.m_bUpdatePalette = FALSE;
    }

    InvalidateRect(MAME32App.m_hWnd, &This.m_DrawRect, FALSE);

    osd_profiler(OSD_PROFILE_BLIT);

    UpdateWindow(MAME32App.m_hWnd);
    osd_profiler(OSD_PROFILE_END);

//    StatusUpdate();

    MAME32App.ProcessMessages();
}

/* control keyboard leds or other indicators */
static void GDI_led_w(int led, int on)
{
//    StatusWrite(led, on & 1);
}

static void GDI_set_gamma(float gamma)
{
    assert(OSDDisplay.set_gamma != 0);

    OSDDisplay.set_gamma(gamma);

    AdjustPalette();
}

static void GDI_set_brightness(int brightness)
{
    assert(OSDDisplay.set_brightness != 0);

    OSDDisplay.set_brightness(brightness);

    AdjustPalette();
}

static void GDI_save_snapshot()
{
    if (This.m_bAviCapture)
    {
        This.m_nAviShowMessage = 10;        /* show message for 10 frames */
		This.m_bAviRun = !This.m_bAviRun;   /* toggle capture on/off */
	}
    else
        Display_WriteBitmap(This.m_pMAMEBitmap, This.m_PalEntries);
}

/***************************************************************************
    Internal functions
 ***************************************************************************/

static struct osd_bitmap* AllocBitmap(struct osd_bitmap *bm, 
	int width, int height, int depth, LPPALETTEENTRY lpPal)
{
	int					i;
	struct GDIBitmap	*gdibm;
	int					palsize;
	int					line_width;
    int					pixel_width, pixel_height;
	unsigned char		*bits;
	int					pos;
	DWORD				*rgbmask;
	HDC					hDC;
    int					safety;

	if (!bm)
	{
		if (Machine->orientation & ORIENTATION_SWAP_XY)
		{
			int     temp;

			temp   = width;
			width  = height;
			height = temp;
		}

		bm = (struct osd_bitmap*)malloc(sizeof(struct osd_bitmap*));
		bm->width = width;
		bm->height = height;
		bm->depth = depth;

		if (!bm)
			return NULL;

		if (bm->width > 64)
		   safety = 8;
		else
		   safety = 0;

		line_width = (((width + safety) * depth / 8) + 7) & ~7;     /* round width to a quadword */

		pixel_width = line_width * 8 / depth;					
		pixel_height = height + safety;

		if (depth <= 8)
			palsize = sizeof(DWORD) * (1 << depth);
		else
			palsize = sizeof(RGBQUAD) * 3;

		gdibm = (struct GDIBitmap*)malloc(sizeof(struct GDIBitmap) + palsize - sizeof(RGBQUAD) + 2048);
		bm->_private = (void*)gdibm;

		bm->line = (unsigned char **)malloc(pixel_height * sizeof(LPVOID));
	}
	else
	{
		gdibm = (struct GDIBitmap*)bm->_private;

		width = bm->width;
		height = bm->height;
		depth = bm->depth;

		safety = -gdibm->bi.bmiHeader.biHeight - bm->height;
		pixel_width = gdibm->bi.bmiHeader.biWidth;
		pixel_height = -gdibm->bi.bmiHeader.biHeight;
		line_width = gdibm->linesize;

		DeleteObject((HGDIOBJ)gdibm->bitmap);
	}

    /* Create BitmapInfo */
    gdibm->bi.bmiHeader.biSize          = sizeof(BITMAPINFOHEADER); 
    gdibm->bi.bmiHeader.biWidth         = pixel_width;
    gdibm->bi.bmiHeader.biHeight        = -pixel_height; /* Negative means "top down" */
    gdibm->bi.bmiHeader.biPlanes        = 1;
    gdibm->bi.bmiHeader.biBitCount      = depth;
    gdibm->bi.bmiHeader.biCompression   = (depth == 16 || depth == 32) ? BI_BITFIELDS : BI_RGB;
    gdibm->bi.bmiHeader.biSizeImage     = 0;
    gdibm->bi.bmiHeader.biXPelsPerMeter = 0;
    gdibm->bi.bmiHeader.biYPelsPerMeter = 0;
    gdibm->bi.bmiHeader.biClrUsed       = 0;
    gdibm->bi.bmiHeader.biClrImportant  = 0;

	rgbmask = (DWORD*)gdibm->bi.bmiColors;
	if (depth == 16)
	{
		((WORD*)(gdibm->bi.bmiColors))[0] = (WORD)(0x1F << 10);
		((WORD*)(gdibm->bi.bmiColors))[1] = (WORD)(0x1F << 5);
		((WORD*)(gdibm->bi.bmiColors))[2] = (WORD)(0x1F << 0);
	}
	else if (depth == 24) // Not actually used in CE..
	{
		((DWORD*)(gdibm->bi.bmiColors))[0] = 0x1F << 10;
		((DWORD*)(gdibm->bi.bmiColors))[1] = 0x1F << 5;
		((DWORD*)(gdibm->bi.bmiColors))[2] = 0x1F << 0;
		rgbmask[0] = 0xFF << 16;
		rgbmask[1] = 0xFF << 8;
		rgbmask[2] = 0xFF << 0;
	}
	else if (depth == 8)
	{
		if (lpPal)
		{
			for (i = 0; i < (1 << depth); i++)
			{
				gdibm->bi.bmiColors[i].rgbRed = lpPal[i].peRed;
				gdibm->bi.bmiColors[i].rgbGreen = lpPal[i].peGreen;
				gdibm->bi.bmiColors[i].rgbBlue = lpPal[i].peBlue;
				gdibm->bi.bmiColors[i].rgbReserved = 0;
			}
		}
		else
			for (i = 0; i < (1 << depth); i++)
				((DWORD*)(gdibm->bi.bmiColors))[i] = 0;
    }

	hDC = GetDC(NULL);
	gdibm->bitmap = CreateDIBSection(hDC, &(gdibm->bi), DIB_RGB_COLORS, &bits, NULL, 0);
    ReleaseDC(NULL, hDC); 

	pos = 0;
	for (i = 0; i < pixel_height; i++)
	{
		bm->line[i] = bits + pos;
		pos += line_width;
	}				

	gdibm->linesize = line_width;
	gdibm->bufsize = line_width * pixel_height;
	
    GDI_clearbitmap(bm);

    return bm;
}

static void SetPen(int pen, unsigned char red, unsigned char green, unsigned char blue)
{
    assert(This.m_nDepth == 8); /* Shouldn't set pen in 16 bit color depth game mode! */

    assert(0 <= pen && pen < OSD_NUMPENS);

    This.m_PalEntries[pen].peRed    = red;
    This.m_PalEntries[pen].peGreen  = green;
    This.m_PalEntries[pen].peBlue   = blue;
    This.m_PalEntries[pen].peFlags  = 0/*PC_NOCOLLAPSE*/; // Not needed in CE

    memcpy(&This.m_AdjustedPalette[pen], &This.m_PalEntries[pen], sizeof(PALETTEENTRY));
    Display_MapColor(&This.m_AdjustedPalette[pen].peRed,
                     &This.m_AdjustedPalette[pen].peGreen,
                     &This.m_AdjustedPalette[pen].peBlue);
}

static void SetPaletteColors(void)
{
    assert(This.m_hPalette != NULL);

    if (This.m_nDepth == 8)
        SetPaletteEntries(This.m_hPalette, 0, OSD_NUMPENS, This.m_AdjustedPalette);    
		
	SelectObject(This.m_hMemDC, (HGDIOBJ)This.m_hOldBitmap);

	AllocBitmap(This.m_pMAMEBitmap, -1, -1, -1, This.m_AdjustedPalette);   // Reallocate the HBITMAP..

	This.m_hBitmap = This.m_pGDIBitmap->bitmap;
	This.m_hOldBitmap = SelectObject(This.m_hMemDC, (HGDIOBJ)This.m_hBitmap);
}

static void AdjustVisibleRect(int xmin, int ymin, int xmax, int ymax)
{
    int temp;
    int w, h;

    if (Machine->orientation & ORIENTATION_SWAP_XY)
    {
        temp = xmin; xmin = ymin; ymin = temp;
        temp = xmax; xmax = ymax; ymax = temp;
        w = Machine->drv->screen_height;
        h = Machine->drv->screen_width;
    }
    else
    {
        w = Machine->drv->screen_width;
        h = Machine->drv->screen_height;
    }

    if (!(Machine->drv->video_attributes & VIDEO_TYPE_VECTOR))
    {
        if (Machine->orientation & ORIENTATION_FLIP_X)
        {
            temp = w - xmin - 1;
            xmin = w - xmax - 1;
            xmax = temp;
        }
        if (Machine->orientation & ORIENTATION_FLIP_Y)
        {
            temp = h - ymin - 1;
            ymin = h - ymax - 1;
            ymax = temp;
        }
    }

//    This.m_VisibleRect.m_Left   = xmin;
//    This.m_VisibleRect.m_Top    = ymin;
//    This.m_VisibleRect.m_Width  = xmax - xmin + 1;
//    This.m_VisibleRect.m_Height = ymax - ymin + 1;
}

static void AdjustPalette(void)
{
    if (This.m_nDepth == 8)
    {
        UINT i;
        
        memcpy(This.m_AdjustedPalette, This.m_PalEntries, OSD_NUMPENS * sizeof(PALETTEENTRY));

        for (i = 0; i < OSD_NUMPENS; i++)
        {
            if (i != This.m_nUIPen) /* Don't map the UI pen. */
                Display_MapColor(&This.m_AdjustedPalette[i].peRed,
                                 &This.m_AdjustedPalette[i].peGreen,
                                 &This.m_AdjustedPalette[i].peBlue);
        }

        This.m_bUpdatePalette = TRUE;
    }
}

static BOOL GDI_OnMessage(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam, LRESULT* pResult)
{
    switch (Msg)
    {
        HANDLE_MESSAGE(hWnd, WM_COMMAND,            OnCommand);
        HANDLE_MESSAGE(hWnd, WM_PAINT,              OnPaint);
//        HANDLE_MESSAGE(hWnd, WM_PALETTECHANGED,     OnPaletteChanged); // Not needed in CE
//        HANDLE_MESSAGE(hWnd, WM_QUERYNEWPALETTE,    OnQueryNewPalette);
//        HANDLE_MESSAGE(hWnd, WM_GETMINMAXINFO,      OnGetMinMaxInfo);	 // Not supported in CE
//        HANDLE_MESSAGE(hWnd, WM_ERASEBKGND,         OnEraseBkgnd);
        HANDLE_MESSAGE(hWnd, WM_SIZE,               OnSize);
        HANDLE_MESSAGE(hWnd, WM_DRAWITEM,           OnDrawItem);
    }
    return FALSE;
}

static void GDI_Refresh()
{
    InvalidateRect(MAME32App.m_hWnd, &This.m_DrawRect, FALSE); 
    UpdateWindow(MAME32App.m_hWnd);
}

static int GDI_GetBlackPen(void)
{
    if (This.m_nDepth == 8)
    {
        int i;
        for (i = 0; i < OSD_NUMPENS; i++)
        {
            if (This.m_PalEntries[i].peRed   == 0
            &&  This.m_PalEntries[i].peGreen == 0
            &&  This.m_PalEntries[i].peBlue  == 0)
            {
                return i;
            }
        }
    }

    return 0;
}

static void GDI_UpdateFPS(BOOL bShow, int nSpeed, int nFPS, int nMachineFPS, int nFrameskip, int nVecUPS)
{
//    StatusUpdateFPS(bShow, nSpeed, nFPS, nMachineFPS, nFrameskip, nVecUPS);
}

static void OnCommand(HWND hWnd, int CmdID, HWND hWndCtl, UINT codeNotify)
{
//    switch (CmdID)
//    {
//        case ID_FILE_EXIT:
//            /* same as double-clicking on main window close box */
//            SendMessage(hWnd, WM_CLOSE, 0, 0);
//        break;
//
//        case ID_ABOUT:
//            ShowWindow(This.m_hWndAbout, SW_SHOW);
//        break;
//    }
}

static void OnPaint(HWND hWnd)
{
    PAINTSTRUCT     ps;
    HPALETTE		hOldPalette;
	BOOL			ok;
	HDC				memdc;
	HBITMAP			bm, oldbm;
	DWORD			ticks;

    BeginPaint(hWnd, &ps);

	/* Update background edges */
//	if (ps.rcPaint.right == This.m_ClientRect.right &&				/* Redraw full window */
//		ps.rcPaint.bottom == This.m_ClientRect.bottom &&
//		(This.m_DrawRect.left > 0 || This.m_DrawRect.bottom > 0))	/* Draw area is smaller than window area */
//	{
//		PatBlt(ps.hdc, 
//			ps.rcPaint.left, 
//			ps.rcPaint.top,
//			RECT_WIDTH(ps.rcPaint),
//			RECT_HEIGHT(ps.rcPaint),
//			BLACKNESS);
//	}

	if (This.m_pMAMEBitmap)
	{
		if (This.m_nScreenDepth == 8)
		{
			hOldPalette = SelectPalette(ps.hdc, This.m_hPalette, FALSE);
			RealizePalette(ps.hdc);
		}

		ticks = GetTickCount();

		ok  = StretchBlt(ps.hdc,
					  This.m_DrawRect.left,
					  This.m_DrawRect.top,
					  RECT_WIDTH(This.m_DrawRect),
					  RECT_HEIGHT(This.m_DrawRect),
					  This.m_hMemDC,
					  0,
					  0,
					  This.m_pMAMEBitmap->width,
					  This.m_pMAMEBitmap->height,
					  SRCCOPY);

		ticks = GetTickCount() - ticks;

		if (This.m_nScreenDepth == 8)
		{
			if (hOldPalette != NULL)
				SelectPalette(ps.hdc, hOldPalette, FALSE);
		}
	}

    EndPaint(hWnd, &ps); 
}

static void OnPaletteChanged(HWND hWnd, HWND hWndPaletteChange)
{
    if (hWnd == hWndPaletteChange) 
        return; 
    
    OnQueryNewPalette(hWnd);
}

static BOOL OnQueryNewPalette(HWND hWnd)
{
    InvalidateRect(hWnd, &This.m_DrawRect, TRUE); 
    UpdateWindow(hWnd); 
    return TRUE;
}

//static void OnGetMinMaxInfo(HWND hWnd, MINMAXINFO* pMinMaxInfo)
//{
//    pMinMaxInfo->ptMaxSize.x      = This.m_nWindowWidth;
//    pMinMaxInfo->ptMaxSize.y      = This.m_nWindowHeight;
//    pMinMaxInfo->ptMaxTrackSize.x = This.m_nWindowWidth;
//    pMinMaxInfo->ptMaxTrackSize.y = This.m_nWindowHeight;
//    pMinMaxInfo->ptMinTrackSize.x = This.m_nWindowWidth;
//    pMinMaxInfo->ptMinTrackSize.y = This.m_nWindowHeight;
//}

static BOOL OnEraseBkgnd(HWND hWnd, HDC hDC)
{
    /* No background erasing is required. */
    return TRUE;
}

static void OnSize(HWND hwnd,UINT state,int cx,int cy)
{
//    StatusWindowSize(state,cx,cy);
}

static void OnDrawItem(HWND hwnd, const DRAWITEMSTRUCT *lpDrawItem)
{
//    StatusDrawItem(lpDrawItem);
}

static LRESULT CALLBACK AboutDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch (Msg)
    {
        case WM_INITDIALOG:
            return 1;

        case WM_COMMAND:
            ShowWindow(hDlg, SW_HIDE);
            return 1;
    }
    return 0;
}

