/***************************************************************************

    M.A.M.E. CE  -  Multiple Arcade Machine Emulator for Windows CE
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse
	Windows CE Port by Benjamin Cooley
    
    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

#ifndef __MAMECE_H__
#define __MAMECE_H__

#include "Options.h"

#define RECT_WIDTH(r)   (r.right  - r.left)
#define RECT_HEIGHT(r)  (r.bottom - r.top)

options_type * GetPlayingGameOptions();

void ErrorMsg(const char* fmt, ...);

TCHAR *WC(const char *str);
TCHAR *WC2(const char *str);
TCHAR *WC3(const char *str);

#define HANDLE_MESSAGE(hwnd, message, fn)                               \
    case (message):                                                     \
    {                                                                   \
        *pResult = HANDLE_##message((hwnd), (wParam), (lParam), (fn));  \
        return TRUE;                                                    \
    }

#define PEEK_MESSAGE(hwnd, message, fn)                                 \
    case (message):                                                     \
    {                                                                   \
        *pResult = HANDLE_##message((hwnd), (wParam), (lParam), (fn));  \
        return FALSE;                                                   \
    }

struct tMAME32App
{
	HINSTANCE				m_hInstance;
    HWND                    m_hwndUI;
    HWND                    m_hWnd;
    const char*             m_Name;
	const char*				m_RootPath;

	int						m_nSpeed;	/* Current game speed percentage */

    BOOL                    m_bIsInitialized;
    BOOL                    m_bIsActive;
    BOOL                    m_bIsPaused;
    BOOL                    m_bDone;
    BOOL                    m_bPaused;

    /*
        TRUE if using m_pTrak for standard analog inputs,
        FALSE if using m_pJoystick for standard analog inputs.
    */
    BOOL                    m_bUseAIMouse; 

    struct OSDDisplay*      m_pDisplay;
    struct OSDSound*        m_pSound;
    struct OSDKeyboard*     m_pKeyboard;
    struct OSDJoystick*     m_pJoystick;
    struct OSDTrak*         m_pTrak;

    struct OSDFMSynth*      m_pFMSynth;

#ifdef MAME_NET
    struct OSDNetwork*      m_pNetwork;
    BOOL                    m_bUseNetwork;
#endif /* MAME_NET */

    HWND                    (*CreateMAMEWindow)(void);
    void                    (*ProcessMessages)(void);
    BOOL                    (*PumpAndReturnMessage)(MSG* pMsg);
    void                    (*Quit)(void);
    BOOL                    (*Done)(void);
    BOOL                    (*Pause)(void);
};

extern TCHAR *WC(const char *str);	// Make wide character

extern void MAME32App_init(options_type *options);

extern struct tMAME32App MAME32App;

typedef struct
{
    BOOL has_roms;
    BOOL has_samples;
    BOOL neogeo_clone;
} game_data_type;

#endif
